@@ -30,8 +30,8 @@ const struct cmd_entry cmd_choose_tree_entry = {
 	.name = "choose-tree",
 	.alias = NULL,
 
-	.args = { "F:f:NO:st:w", 0, 1 },
-	.usage = "[-Nsw] [-F format] [-f filter] [-O sort-order] "
+	.args = { "F:Gf:NO:st:wZ", 0, 1 },
+	.usage = "[-GNsw] [-F format] [-f filter] [-O sort-order] "
 	         CMD_TARGET_PANE_USAGE,
 
 	.target = { 't', CMD_FIND_PANE, 0 },
@@ -44,7 +44,7 @@ const struct cmd_entry cmd_choose_client_entry = {
 	.name = "choose-client",
 	.alias = NULL,
 
-	.args = { "F:f:NO:t:", 0, 1 },
+	.args = { "F:f:NO:t:Z", 0, 1 },
 	.usage = "[-N] [-F format] [-f filter] [-O sort-order] "
 	         CMD_TARGET_PANE_USAGE,
 
@@ -58,7 +58,7 @@ const struct cmd_entry cmd_choose_buffer_entry = {
 	.name = "choose-buffer",
 	.alias = NULL,
 
-	.args = { "F:f:NO:t:", 0, 1 },
+	.args = { "F:f:NO:t:Z", 0, 1 },
 	.usage = "[-N] [-F format] [-f filter] [-O sort-order] "
 	         CMD_TARGET_PANE_USAGE,
 
