@@ -1,7 +1,7 @@
-/* $Id$ */
+/* $OpenBSD$ */
 
 /*
- * Copyright (c) 2008 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2008 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -18,62 +18,61 @@
 
 #include <sys/types.h>
 
-#include <pwd.h>
-#include <string.h>
-#include <unistd.h>
-
 #include "tmux.h"
 
 /*
  * Lock commands.
  */
 
-enum cmd_retval	 cmd_lock_server_exec(struct cmd *, struct cmd_q *);
+static enum cmd_retval	cmd_lock_server_exec(struct cmd *, struct cmdq_item *);
 
 const struct cmd_entry cmd_lock_server_entry = {
-	"lock-server", "lock",
-	"", 0, 0,
-	"",
-	0,
-	NULL,
-	cmd_lock_server_exec
+	.name = "lock-server",
+	.alias = "lock",
+
+	.args = { "", 0, 0 },
+	.usage = "",
+
+	.flags = CMD_AFTERHOOK,
+	.exec = cmd_lock_server_exec
 };
 
 const struct cmd_entry cmd_lock_session_entry = {
-	"lock-session", "locks",
-	"t:", 0, 0,
-	CMD_TARGET_SESSION_USAGE,
-	0,
-	NULL,
-	cmd_lock_server_exec
+	.name = "lock-session",
+	.alias = "locks",
+
+	.args = { "t:", 0, 0 },
+	.usage = CMD_TARGET_SESSION_USAGE,
+
+	.target = { 't', CMD_FIND_SESSION, 0 },
+
+	.flags = CMD_AFTERHOOK,
+	.exec = cmd_lock_server_exec
 };
 
 const struct cmd_entry cmd_lock_client_entry = {
-	"lock-client", "lockc",
-	"t:", 0, 0,
-	CMD_TARGET_CLIENT_USAGE,
-	0,
-	NULL,
-	cmd_lock_server_exec
+	.name = "lock-client",
+	.alias = "lockc",
+
+	.args = { "t:", 0, 0 },
+	.usage = CMD_TARGET_CLIENT_USAGE,
+
+	.flags = CMD_AFTERHOOK,
+	.exec = cmd_lock_server_exec
 };
 
-enum cmd_retval
-cmd_lock_server_exec(struct cmd *self, unused struct cmd_q *cmdq)
+static enum cmd_retval
+cmd_lock_server_exec(struct cmd *self, struct cmdq_item *item)
 {
 	struct args	*args = self->args;
 	struct client	*c;
-	struct session	*s;
 
 	if (self->entry == &cmd_lock_server_entry)
 		server_lock();
-	else if (self->entry == &cmd_lock_session_entry) {
-		s = cmd_find_session(cmdq, args_get(args, 't'), 0);
-		if (s == NULL)
-			return (CMD_RETURN_ERROR);
-		server_lock_session(s);
-	} else {
-		c = cmd_find_client(cmdq, args_get(args, 't'), 0);
-		if (c == NULL)
+	else if (self->entry == &cmd_lock_session_entry)
+		server_lock_session(item->target.s);
+	else {
+		if ((c = cmd_find_client(item, args_get(args, 't'), 0)) == NULL)
 			return (CMD_RETURN_ERROR);
 		server_lock_client(c);
 	}
