@@ -27,9 +27,10 @@
 static const char *window_copy_key_table(struct window_pane *);
 static void	window_copy_command(struct window_pane *, struct client *,
 		    struct session *, struct args *, struct mouse_event *);
-static struct screen *window_copy_init(struct window_pane *);
+static struct screen *window_copy_init(struct window_pane *,
+		    struct cmd_find_state *, struct args *);
 static void	window_copy_free(struct window_pane *);
-static int	window_copy_pagedown(struct window_pane *, int);
+static int	window_copy_pagedown(struct window_pane *, int, int);
 static void	window_copy_next_paragraph(struct window_pane *);
 static void	window_copy_previous_paragraph(struct window_pane *);
 static void	window_copy_resize(struct window_pane *, u_int, u_int);
@@ -56,16 +57,17 @@ static void	window_copy_move_right(struct screen *, u_int *, u_int *);
 static int	window_copy_is_lowercase(const char *);
 static int	window_copy_search_jump(struct window_pane *, struct grid *,
 		    struct grid *, u_int, u_int, u_int, int, int, int);
-static int	window_copy_search(struct window_pane *, int, int);
-static int	window_copy_search_up(struct window_pane *, int);
-static int	window_copy_search_down(struct window_pane *, int);
+static int	window_copy_search(struct window_pane *, int);
+static int	window_copy_search_up(struct window_pane *);
+static int	window_copy_search_down(struct window_pane *);
 static void	window_copy_goto_line(struct window_pane *, const char *);
 static void	window_copy_update_cursor(struct window_pane *, u_int, u_int);
 static void	window_copy_start_selection(struct window_pane *);
 static int	window_copy_adjust_selection(struct window_pane *, u_int *,
 		    u_int *);
+static int	window_copy_set_selection(struct window_pane *, int);
 static int	window_copy_update_selection(struct window_pane *, int);
-static void	window_copy_synchronize_cursor(struct window_pane *wp);
+static void	window_copy_synchronize_cursor(struct window_pane *);
 static void    *window_copy_get_selection(struct window_pane *, size_t *);
 static void	window_copy_copy_buffer(struct window_pane *, const char *,
 		    void *, size_t);
@@ -90,8 +92,8 @@ static void	window_copy_cursor_up(struct window_pane *, int);
 static void	window_copy_cursor_down(struct window_pane *, int);
 static void	window_copy_cursor_jump(struct window_pane *);
 static void	window_copy_cursor_jump_back(struct window_pane *);
-static void	window_copy_cursor_jump_to(struct window_pane *, int);
-static void	window_copy_cursor_jump_to_back(struct window_pane *, int);
+static void	window_copy_cursor_jump_to(struct window_pane *);
+static void	window_copy_cursor_jump_to_back(struct window_pane *);
 static void	window_copy_cursor_next_word(struct window_pane *,
 		    const char *);
 static void	window_copy_cursor_next_word_end(struct window_pane *,
@@ -150,12 +152,12 @@ struct window_copy_mode_data {
 	struct screen	*backing;
 	int		 backing_written; /* backing display started */
 
-	u_int		 oy; /* number of lines scrolled up */
+	u_int		 oy;		/* number of lines scrolled up */
 
-	u_int		 selx; /* beginning of selection */
+	u_int		 selx;		/* beginning of selection */
 	u_int		 sely;
 
-	u_int		 endselx; /* end of selection */
+	u_int		 endselx;	/* end of selection */
 	u_int		 endsely;
 
 	enum {
@@ -164,14 +166,20 @@ struct window_copy_mode_data {
 		CURSORDRAG_SEL,		/* start is synchronized with cursor */
 	} cursordrag;
 
+	int		 modekeys;
+	enum {
+		LINE_SEL_NONE,
+		LINE_SEL_LEFT_RIGHT,
+		LINE_SEL_RIGHT_LEFT,
+	} lineflag;			/* line selection mode */
 	int		 rectflag;	/* in rectangle copy mode? */
 	int		 scroll_exit;	/* exit on scroll to end? */
 
 	u_int		 cx;
 	u_int		 cy;
 
-	u_int		 lastcx; /* position in last line w/ content */
-	u_int		 lastsx; /* size of last line w/ content */
+	u_int		 lastcx; 	/* position in last line w/ content */
+	u_int		 lastsx;	/* size of last line w/ content */
 
 	int		 searchtype;
 	char		*searchstr;
@@ -187,29 +195,24 @@ struct window_copy_mode_data {
 };
 
 static struct screen *
-window_copy_init(struct window_pane *wp)
+window_copy_init(struct window_pane *wp, __unused struct cmd_find_state *fs,
+    __unused struct args *args)
 {
 	struct window_copy_mode_data	*data;
 	struct screen			*s;
 
-	wp->modedata = data = xmalloc(sizeof *data);
-
-	data->oy = 0;
-	data->cx = 0;
-	data->cy = 0;
+	wp->modedata = data = xcalloc(1, sizeof *data);
 
 	data->cursordrag = CURSORDRAG_NONE;
+	data->lineflag = LINE_SEL_NONE;
 
-	data->lastcx = 0;
-	data->lastsx = 0;
-
-	data->backing_written = 0;
-
-	data->rectflag = 0;
-	data->scroll_exit = 0;
-
-	data->searchtype = WINDOW_COPY_OFF;
-	data->searchstr = NULL;
+	if (wp->searchstr != NULL) {
+		data->searchtype = WINDOW_COPY_SEARCHUP;
+		data->searchstr = xstrdup(wp->searchstr);
+	} else {
+		data->searchtype = WINDOW_COPY_OFF;
+		data->searchstr = NULL;
+	}
 	data->searchmark = NULL;
 	data->searchx = data->searchy = data->searcho = -1;
 
@@ -221,7 +224,7 @@ window_copy_init(struct window_pane *wp)
 
 	s = &data->screen;
 	screen_init(s, screen_size_x(&wp->base), screen_size_y(&wp->base), 0);
-	s->sel.modekeys = options_get_number(wp->window->options, "mode-keys");
+	data->modekeys = options_get_number(wp->window->options, "mode-keys");
 
 	data->backing = NULL;
 
@@ -316,7 +319,7 @@ window_copy_vadd(struct window_pane *wp, const char *fmt, va_list ap)
 		 * (so it's on a new line).
 		 */
 		screen_write_carriagereturn(&back_ctx);
-		screen_write_linefeed(&back_ctx, 0);
+		screen_write_linefeed(&back_ctx, 0, 8);
 	} else
 		data->backing_written = 1;
 	old_cy = backing->cy;
@@ -350,7 +353,7 @@ window_copy_pageup(struct window_pane *wp, int half_page)
 	oy = screen_hsize(data->backing) + data->cy - data->oy;
 	ox = window_copy_find_length(wp, oy);
 
-	if (s->sel.lineflag == LINE_SEL_LEFT_RIGHT && oy == data->sely)
+	if (data->lineflag == LINE_SEL_LEFT_RIGHT && oy == data->sely)
 		window_copy_other_end(wp);
 
 	if (data->cx != ox) {
@@ -372,7 +375,7 @@ window_copy_pageup(struct window_pane *wp, int half_page)
 	else
 		data->oy += n;
 
-	if (!data->screen.sel.flag || !data->rectflag) {
+	if (data->screen.sel == NULL || !data->rectflag) {
 		py = screen_hsize(data->backing) + data->cy - data->oy;
 		px = window_copy_find_length(wp, py);
 		if ((data->cx >= data->lastsx && data->cx != px) ||
@@ -385,7 +388,7 @@ window_copy_pageup(struct window_pane *wp, int half_page)
 }
 
 static int
-window_copy_pagedown(struct window_pane *wp, int half_page)
+window_copy_pagedown(struct window_pane *wp, int half_page, int scroll_exit)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*s = &data->screen;
@@ -394,7 +397,7 @@ window_copy_pagedown(struct window_pane *wp, int half_page)
 	oy = screen_hsize(data->backing) + data->cy - data->oy;
 	ox = window_copy_find_length(wp, oy);
 
-	if (s->sel.lineflag == LINE_SEL_RIGHT_LEFT && oy == data->sely)
+	if (data->lineflag == LINE_SEL_RIGHT_LEFT && oy == data->sely)
 		window_copy_other_end(wp);
 
 	if (data->cx != ox) {
@@ -416,7 +419,7 @@ window_copy_pagedown(struct window_pane *wp, int half_page)
 	else
 		data->oy -= n;
 
-	if (!data->screen.sel.flag || !data->rectflag) {
+	if (data->screen.sel == NULL || !data->rectflag) {
 		py = screen_hsize(data->backing) + data->cy - data->oy;
 		px = window_copy_find_length(wp, py);
 		if ((data->cx >= data->lastsx && data->cx != px) ||
@@ -424,7 +427,7 @@ window_copy_pagedown(struct window_pane *wp, int half_page)
 			window_copy_cursor_end_of_line(wp);
 	}
 
-	if (data->scroll_exit && data->oy == 0)
+	if (scroll_exit && data->oy == 0)
 		return (1);
 	window_copy_update_selection(wp, 1);
 	window_copy_redraw_screen(wp);
@@ -517,7 +520,7 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 	struct screen			*sn = &data->screen;
 	const char			*command, *argument, *ws;
 	u_int				 np = wp->modeprefix;
-	int				 cancel = 0, redraw = 0;
+	int				 cancel = 0, redraw = 0, scroll_exit;
 	char				 prefix;
 
 	if (args->argc == 0)
@@ -547,7 +550,7 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 			if (m != NULL)
 				window_copy_start_drag(c, m);
 			else {
-				sn->sel.lineflag = LINE_SEL_NONE;
+				data->lineflag = LINE_SEL_NONE;
 				window_copy_start_selection(wp);
 				redraw = 1;
 			}
@@ -622,9 +625,14 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 		}
 		if (strcmp(command, "end-of-line") == 0)
 			window_copy_cursor_end_of_line(wp);
-		if (strcmp(command, "halfpage-down") == 0) {
+		if (strcmp(command, "halfpage-down") == 0 ||
+		    strcmp(command, "halfpage-down-and-cancel") == 0) {
+			if (strcmp(command, "halfpage-down-and-cancel") == 0)
+				scroll_exit = 1;
+			else
+				scroll_exit = data->scroll_exit;
 			for (; np != 0; np--) {
-				if (window_copy_pagedown(wp, 1)) {
+				if (window_copy_pagedown(wp, 1, scroll_exit)) {
 					cancel = 1;
 					break;
 				}
@@ -660,11 +668,11 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 				break;
 			case WINDOW_COPY_JUMPTOFORWARD:
 				for (; np != 0; np--)
-					window_copy_cursor_jump_to(wp, 1);
+					window_copy_cursor_jump_to(wp);
 				break;
 			case WINDOW_COPY_JUMPTOBACKWARD:
 				for (; np != 0; np--)
-					window_copy_cursor_jump_to_back(wp, 1);
+					window_copy_cursor_jump_to_back(wp);
 				break;
 			}
 		}
@@ -680,11 +688,11 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 				break;
 			case WINDOW_COPY_JUMPTOFORWARD:
 				for (; np != 0; np--)
-					window_copy_cursor_jump_to_back(wp, 1);
+					window_copy_cursor_jump_to_back(wp);
 				break;
 			case WINDOW_COPY_JUMPTOBACKWARD:
 				for (; np != 0; np--)
-					window_copy_cursor_jump_to(wp, 1);
+					window_copy_cursor_jump_to(wp);
 				break;
 			}
 		}
@@ -720,9 +728,14 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 			if ((np % 2) != 0)
 				window_copy_other_end(wp);
 		}
-		if (strcmp(command, "page-down") == 0) {
+		if (strcmp(command, "page-down") == 0 ||
+		    strcmp(command, "page-down-and-cancel") == 0) {
+			if (strcmp(command, "page-down-and-cancel") == 0)
+				scroll_exit = 1;
+			else
+				scroll_exit = data->scroll_exit;
 			for (; np != 0; np--) {
-				if (window_copy_pagedown(wp, 0)) {
+				if (window_copy_pagedown(wp, 0, scroll_exit)) {
 					cancel = 1;
 					break;
 				}
@@ -746,13 +759,18 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 				window_copy_cursor_previous_word(wp, ws);
 		}
 		if (strcmp(command, "rectangle-toggle") == 0) {
-			sn->sel.lineflag = LINE_SEL_NONE;
+			data->lineflag = LINE_SEL_NONE;
 			window_copy_rectangle_toggle(wp);
 		}
-		if (strcmp(command, "scroll-down") == 0) {
+		if (strcmp(command, "scroll-down") == 0 ||
+		    strcmp(command, "scroll-down-and-cancel") == 0) {
+			if (strcmp(command, "scroll-down-and-cancel") == 0)
+				scroll_exit = 1;
+			else
+				scroll_exit = data->scroll_exit;
 			for (; np != 0; np--)
 				window_copy_cursor_down(wp, 1);
-			if (data->scroll_exit && data->oy == 0)
+			if (scroll_exit && data->oy == 0)
 				cancel = 1;
 		}
 		if (strcmp(command, "scroll-up") == 0) {
@@ -762,23 +780,23 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 		if (strcmp(command, "search-again") == 0) {
 			if (data->searchtype == WINDOW_COPY_SEARCHUP) {
 				for (; np != 0; np--)
-					window_copy_search_up(wp, 1);
+					window_copy_search_up(wp);
 			} else if (data->searchtype == WINDOW_COPY_SEARCHDOWN) {
 				for (; np != 0; np--)
-					window_copy_search_down(wp, 1);
+					window_copy_search_down(wp);
 			}
 		}
 		if (strcmp(command, "search-reverse") == 0) {
 			if (data->searchtype == WINDOW_COPY_SEARCHUP) {
 				for (; np != 0; np--)
-					window_copy_search_down(wp, 1);
+					window_copy_search_down(wp);
 			} else if (data->searchtype == WINDOW_COPY_SEARCHDOWN) {
 				for (; np != 0; np--)
-					window_copy_search_up(wp, 1);
+					window_copy_search_up(wp);
 			}
 		}
 		if (strcmp(command, "select-line") == 0) {
-			sn->sel.lineflag = LINE_SEL_LEFT_RIGHT;
+			data->lineflag = LINE_SEL_LEFT_RIGHT;
 			data->rectflag = 0;
 			window_copy_cursor_start_of_line(wp);
 			window_copy_start_selection(wp);
@@ -788,7 +806,7 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 			redraw = 1;
 		}
 		if (strcmp(command, "select-word") == 0) {
-			sn->sel.lineflag = LINE_SEL_LEFT_RIGHT;
+			data->lineflag = LINE_SEL_LEFT_RIGHT;
 			data->rectflag = 0;
 			ws = options_get_string(s->options, "word-separators");
 			window_copy_cursor_previous_word(wp, ws);
@@ -834,27 +852,27 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 			data->jumptype = WINDOW_COPY_JUMPTOBACKWARD;
 			data->jumpchar = *argument;
 			for (; np != 0; np--)
-				window_copy_cursor_jump_to_back(wp, 1);
+				window_copy_cursor_jump_to_back(wp);
 		}
 		if (strcmp(command, "jump-to-forward") == 0) {
 			data->jumptype = WINDOW_COPY_JUMPTOFORWARD;
 			data->jumpchar = *argument;
 			for (; np != 0; np--)
-				window_copy_cursor_jump_to(wp, 1);
+				window_copy_cursor_jump_to(wp);
 		}
 		if (strcmp(command, "search-backward") == 0) {
 			data->searchtype = WINDOW_COPY_SEARCHUP;
 			free(data->searchstr);
 			data->searchstr = xstrdup(argument);
 			for (; np != 0; np--)
-				window_copy_search_up(wp, 1);
+				window_copy_search_up(wp);
 		}
 		if (strcmp(command, "search-forward") == 0) {
 			data->searchtype = WINDOW_COPY_SEARCHDOWN;
 			free(data->searchstr);
 			data->searchstr = xstrdup(argument);
 			for (; np != 0; np--)
-				window_copy_search_down(wp, 1);
+				window_copy_search_down(wp);
 		}
 		if (strcmp(command, "search-backward-incremental") == 0) {
 			prefix = *argument++;
@@ -876,7 +894,7 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 				data->searchtype = WINDOW_COPY_SEARCHUP;
 				free(data->searchstr);
 				data->searchstr = xstrdup(argument);
-				if (!window_copy_search_up(wp, 1)) {
+				if (!window_copy_search_up(wp)) {
 					window_copy_clear_marks(wp);
 					redraw = 1;
 				}
@@ -884,7 +902,7 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 				data->searchtype = WINDOW_COPY_SEARCHDOWN;
 				free(data->searchstr);
 				data->searchstr = xstrdup(argument);
-				if (!window_copy_search_down(wp, 1)) {
+				if (!window_copy_search_down(wp)) {
 					window_copy_clear_marks(wp);
 					redraw = 1;
 				}
@@ -910,7 +928,7 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 				data->searchtype = WINDOW_COPY_SEARCHDOWN;
 				free(data->searchstr);
 				data->searchstr = xstrdup(argument);
-				if (!window_copy_search_down(wp, 1)) {
+				if (!window_copy_search_down(wp)) {
 					window_copy_clear_marks(wp);
 					redraw = 1;
 				}
@@ -918,7 +936,7 @@ window_copy_command(struct window_pane *wp, struct client *c, struct session *s,
 				data->searchtype = WINDOW_COPY_SEARCHUP;
 				free(data->searchstr);
 				data->searchstr = xstrdup(argument);
-				if (!window_copy_search_up(wp, 1)) {
+				if (!window_copy_search_up(wp)) {
 					window_copy_clear_marks(wp);
 					redraw = 1;
 				}
@@ -948,18 +966,22 @@ window_copy_scroll_to(struct window_pane *wp, u_int px, u_int py)
 
 	data->cx = px;
 
-	gap = gd->sy / 4;
-	if (py < gd->sy) {
-		offset = 0;
-		data->cy = py;
-	} else if (py > gd->hsize + gd->sy - gap) {
-		offset = gd->hsize;
-		data->cy = py - gd->hsize;
-	} else {
-		offset = py + gap - gd->sy;
-		data->cy = py - offset;
+	if (py >= gd->hsize - data->oy && py < gd->hsize - data->oy + gd->sy)
+		data->cy = py - (gd->hsize - data->oy);
+	else {
+		gap = gd->sy / 4;
+		if (py < gd->sy) {
+			offset = 0;
+			data->cy = py;
+		} else if (py > gd->hsize + gd->sy - gap) {
+			offset = gd->hsize;
+			data->cy = py - gd->hsize;
+		} else {
+			offset = py + gap - gd->sy;
+			data->cy = py - offset;
+		}
+		data->oy = gd->hsize - offset;
 	}
-	data->oy = gd->hsize - offset;
 
 	window_copy_update_selection(wp, 1);
 	window_copy_redraw_screen(wp);
@@ -994,7 +1016,7 @@ window_copy_search_lr(struct grid *gd,
 	int	matched;
 
 	for (ax = first; ax < last; ax++) {
-		if (ax + sgd->sx >= gd->sx)
+		if (ax + sgd->sx > gd->sx)
 			break;
 		for (bx = 0; bx < sgd->sx; bx++) {
 			px = ax + bx;
@@ -1121,11 +1143,10 @@ window_copy_search_jump(struct window_pane *wp, struct grid *gd,
 
 /*
  * Search in for text searchstr. If direction is 0 then search up, otherwise
- * down. If moveflag is 0 then look for string at the current cursor position
- * as well.
+ * down.
  */
 static int
-window_copy_search(struct window_pane *wp, int direction, int moveflag)
+window_copy_search(struct window_pane *wp, int direction)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*s = data->backing, ss;
@@ -1134,6 +1155,9 @@ window_copy_search(struct window_pane *wp, int direction, int moveflag)
 	u_int				 fx, fy, endline;
 	int				 wrapflag, cis, found;
 
+	free(wp->searchstr);
+	wp->searchstr = xstrdup(data->searchstr);
+
 	fx = data->cx;
 	fy = screen_hsize(data->backing) - data->oy + data->cy;
 
@@ -1142,12 +1166,10 @@ window_copy_search(struct window_pane *wp, int direction, int moveflag)
 	screen_write_nputs(&ctx, -1, &grid_default_cell, "%s", data->searchstr);
 	screen_write_stop(&ctx);
 
-	if (moveflag) {
-		if (direction)
-			window_copy_move_right(s, &fx, &fy);
-		else
-			window_copy_move_left(s, &fx, &fy);
-	}
+	if (direction)
+		window_copy_move_right(s, &fx, &fy);
+	else
+		window_copy_move_left(s, &fx, &fy);
 	window_copy_clear_selection(wp);
 
 	wrapflag = options_get_number(wp->window->options, "wrap-search");
@@ -1233,15 +1255,15 @@ window_copy_clear_marks(struct window_pane *wp)
 }
 
 static int
-window_copy_search_up(struct window_pane *wp, int moveflag)
+window_copy_search_up(struct window_pane *wp)
 {
-	return (window_copy_search(wp, 0, moveflag));
+	return (window_copy_search(wp, 0));
 }
 
 static int
-window_copy_search_down(struct window_pane *wp, int moveflag)
+window_copy_search_down(struct window_pane *wp)
 {
-	return (window_copy_search(wp, 1, moveflag));
+	return (window_copy_search(wp, 1));
 }
 
 static void
@@ -1408,7 +1430,6 @@ static void
 window_copy_start_selection(struct window_pane *wp)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
-	struct screen			*s = &data->screen;
 
 	data->selx = data->cx;
 	data->sely = screen_hsize(data->backing) + data->cy - data->oy;
@@ -1418,8 +1439,7 @@ window_copy_start_selection(struct window_pane *wp)
 
 	data->cursordrag = CURSORDRAG_ENDSEL;
 
-	s->sel.flag = 1;
-	window_copy_update_selection(wp, 1);
+	window_copy_set_selection(wp, 1);
 }
 
 static int
@@ -1450,12 +1470,23 @@ window_copy_adjust_selection(struct window_pane *wp, u_int *selx, u_int *sely)
 	}
 
 	*selx = sx;
-	*sely = screen_hsize(s) + sy;
+	*sely = sy;
 	return (relpos);
 }
 
 static int
 window_copy_update_selection(struct window_pane *wp, int may_redraw)
+{
+	struct window_copy_mode_data	*data = wp->modedata;
+	struct screen			*s = &data->screen;
+
+	if (s->sel == NULL && data->lineflag == LINE_SEL_NONE)
+		return (0);
+	return (window_copy_set_selection(wp, may_redraw));
+}
+
+static int
+window_copy_set_selection(struct window_pane *wp, int may_redraw)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*s = &data->screen;
@@ -1464,9 +1495,6 @@ window_copy_update_selection(struct window_pane *wp, int may_redraw)
 	u_int				 sx, sy, cy, endsx, endsy;
 	int				 startrelpos, endrelpos;
 
-	if (!s->sel.flag && s->sel.lineflag == LINE_SEL_NONE)
-		return (0);
-
 	window_copy_synchronize_cursor(wp);
 
 	/* Adjust the selection. */
@@ -1489,7 +1517,8 @@ window_copy_update_selection(struct window_pane *wp, int may_redraw)
 	/* Set colours and selection. */
 	style_apply(&gc, oo, "mode-style");
 	gc.flags |= GRID_FLAG_NOPALETTE;
-	screen_set_selection(s, sx, sy, endsx, endsy, data->rectflag, &gc);
+	screen_set_selection(s, sx, sy, endsx, endsy, data->rectflag,
+	    data->modekeys, &gc);
 
 	if (data->rectflag && may_redraw) {
 		/*
@@ -1498,10 +1527,17 @@ window_copy_update_selection(struct window_pane *wp, int may_redraw)
 		 * of lines, and redraw just past that in both directions
 		 */
 		cy = data->cy;
-		if (sy < cy)
-			window_copy_redraw_lines(wp, sy, cy - sy + 1);
-		else
-			window_copy_redraw_lines(wp, cy, sy - cy + 1);
+		if (data->cursordrag == CURSORDRAG_ENDSEL) {
+			if (sy < cy)
+				window_copy_redraw_lines(wp, sy, cy - sy + 1);
+			else
+				window_copy_redraw_lines(wp, cy, sy - cy + 1);
+		} else {
+			if (endsy < cy)
+				window_copy_redraw_lines(wp, endsy, cy - endsy + 1);
+			else
+				window_copy_redraw_lines(wp, cy, endsy - cy + 1);
+		}
 	}
 
 	return (1);
@@ -1515,10 +1551,10 @@ window_copy_get_selection(struct window_pane *wp, size_t *len)
 	char				*buf;
 	size_t				 off;
 	u_int				 i, xx, yy, sx, sy, ex, ey, ey_last;
-	u_int				 firstsx, lastex, restex, restsx;
+	u_int				 firstsx, lastex, restex, restsx, selx;
 	int				 keys;
 
-	if (!s->sel.flag && s->sel.lineflag == LINE_SEL_NONE)
+	if (data->screen.sel == NULL && data->lineflag == LINE_SEL_NONE)
 		return (NULL);
 
 	buf = xmalloc(1);
@@ -1566,7 +1602,11 @@ window_copy_get_selection(struct window_pane *wp, size_t *len)
 		 * Need to ignore the column with the cursor in it, which for
 		 * rectangular copy means knowing which side the cursor is on.
 		 */
-		if (data->selx < data->cx) {
+		if (data->cursordrag == CURSORDRAG_ENDSEL)
+			selx = data->selx;
+		else
+			selx = data->endselx;
+		if (selx < data->cx) {
 			/* Selection start is on the left. */
 			if (keys == MODEKEY_EMACS) {
 				lastex = data->cx;
@@ -1576,12 +1616,12 @@ window_copy_get_selection(struct window_pane *wp, size_t *len)
 				lastex = data->cx + 1;
 				restex = data->cx + 1;
 			}
-			firstsx = data->selx;
-			restsx = data->selx;
+			firstsx = selx;
+			restsx = selx;
 		} else {
 			/* Cursor is on the left. */
-			lastex = data->selx + 1;
-			restex = data->selx + 1;
+			lastex = selx + 1;
+			restex = selx + 1;
 			firstsx = data->cx;
 			restsx = data->cx;
 		}
@@ -1619,10 +1659,11 @@ window_copy_copy_buffer(struct window_pane *wp, const char *bufname, void *buf,
 {
 	struct screen_write_ctx	ctx;
 
-	if (options_get_number(global_options, "set-clipboard")) {
+	if (options_get_number(global_options, "set-clipboard") != 0) {
 		screen_write_start(&ctx, wp, NULL);
 		screen_write_setselection(&ctx, buf, len);
 		screen_write_stop(&ctx);
+		notify_pane("pane-set-clipboard", wp);
 	}
 
 	if (paste_set(buf, len, bufname, NULL) != 0)
@@ -1643,7 +1684,7 @@ window_copy_copy_pipe(struct window_pane *wp, struct session *s,
 		return;
 	expanded = format_single(NULL, arg, NULL, s, NULL, wp);
 
-	job = job_run(expanded, s, NULL, NULL, NULL, NULL, NULL);
+	job = job_run(expanded, s, NULL, NULL, NULL, NULL, NULL, JOB_NOWAIT);
 	bufferevent_write(job->event, buf, len);
 
 	free(expanded);
@@ -1676,10 +1717,11 @@ window_copy_append_selection(struct window_pane *wp, const char *bufname)
 	if (buf == NULL)
 		return;
 
-	if (options_get_number(global_options, "set-clipboard")) {
+	if (options_get_number(global_options, "set-clipboard") != 0) {
 		screen_write_start(&ctx, wp, NULL);
 		screen_write_setselection(&ctx, buf, len);
 		screen_write_stop(&ctx);
+		notify_pane("pane-set-clipboard", wp);
 	}
 
 	if (bufname == NULL || *bufname == '\0')
@@ -1716,7 +1758,7 @@ window_copy_copy_line(struct window_pane *wp, char **buf, size_t *off, u_int sy,
 	 * Work out if the line was wrapped at the screen edge and all of it is
 	 * on screen.
 	 */
-	gl = &gd->linedata[sy];
+	gl = grid_get_line(gd, sy);
 	if (gl->flags & GRID_LINE_WRAPPED && gl->cellsize <= gd->sx)
 		wrapped = 1;
 
@@ -1804,7 +1846,7 @@ window_copy_find_length(struct window_pane *wp, u_int py)
 	 * width of the grid, and screen_write_copy treats them as spaces, so
 	 * ignore them here too.
 	 */
-	px = s->grid->linedata[py].cellsize;
+	px = grid_get_line(s->grid, py)->cellsize;
 	if (px > screen_size_x(s))
 		px = screen_size_x(s);
 	while (px > 0) {
@@ -1821,14 +1863,13 @@ window_copy_cursor_start_of_line(struct window_pane *wp)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*back_s = data->backing;
-	struct screen			*s = &data->screen;
 	struct grid			*gd = back_s->grid;
 	u_int				 py;
 
-	if (data->cx == 0 && s->sel.lineflag == LINE_SEL_NONE) {
+	if (data->cx == 0 && data->lineflag == LINE_SEL_NONE) {
 		py = screen_hsize(back_s) + data->cy - data->oy;
 		while (py > 0 &&
-		    gd->linedata[py-1].flags & GRID_LINE_WRAPPED) {
+		    grid_get_line(gd, py - 1)->flags & GRID_LINE_WRAPPED) {
 			window_copy_cursor_up(wp, 0);
 			py = screen_hsize(back_s) + data->cy - data->oy;
 		}
@@ -1866,22 +1907,24 @@ window_copy_cursor_end_of_line(struct window_pane *wp)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*back_s = data->backing;
-	struct screen			*s = &data->screen;
 	struct grid			*gd = back_s->grid;
+	struct grid_line		*gl;
 	u_int				 px, py;
 
 	py = screen_hsize(back_s) + data->cy - data->oy;
 	px = window_copy_find_length(wp, py);
 
-	if (data->cx == px && s->sel.lineflag == LINE_SEL_NONE) {
-		if (data->screen.sel.flag && data->rectflag)
+	if (data->cx == px && data->lineflag == LINE_SEL_NONE) {
+		if (data->screen.sel != NULL && data->rectflag)
 			px = screen_size_x(back_s);
-		if (gd->linedata[py].flags & GRID_LINE_WRAPPED) {
-			while (py < gd->sy + gd->hsize &&
-			    gd->linedata[py].flags & GRID_LINE_WRAPPED) {
+		gl = grid_get_line(gd, py);
+		if (gl->flags & GRID_LINE_WRAPPED) {
+			while (py < gd->sy + gd->hsize) {
+				gl = grid_get_line(gd, py);
+				if (~gl->flags & GRID_LINE_WRAPPED)
+					break;
 				window_copy_cursor_down(wp, 0);
-				py = screen_hsize(back_s)
-				     + data->cy - data->oy;
+				py = screen_hsize(back_s) + data->cy - data->oy;
 			}
 			px = window_copy_find_length(wp, py);
 		}
@@ -1899,13 +1942,13 @@ window_copy_other_end(struct window_pane *wp)
 	struct screen			*s = &data->screen;
 	u_int				 selx, sely, cy, yy, hsize;
 
-	if (!s->sel.flag && s->sel.lineflag == LINE_SEL_NONE)
+	if (s->sel == NULL && data->lineflag == LINE_SEL_NONE)
 		return;
 
-	if (s->sel.lineflag == LINE_SEL_LEFT_RIGHT)
-		s->sel.lineflag = LINE_SEL_RIGHT_LEFT;
-	else if (s->sel.lineflag == LINE_SEL_RIGHT_LEFT)
-		s->sel.lineflag = LINE_SEL_LEFT_RIGHT;
+	if (data->lineflag == LINE_SEL_LEFT_RIGHT)
+		data->lineflag = LINE_SEL_RIGHT_LEFT;
+	else if (data->lineflag == LINE_SEL_RIGHT_LEFT)
+		data->lineflag = LINE_SEL_LEFT_RIGHT;
 
 	switch (data->cursordrag) {
 		case CURSORDRAG_NONE:
@@ -1977,7 +2020,7 @@ window_copy_cursor_right(struct window_pane *wp)
 
 	py = screen_hsize(data->backing) + data->cy - data->oy;
 	yy = screen_hsize(data->backing) + screen_size_y(data->backing) - 1;
-	if (data->screen.sel.flag && data->rectflag)
+	if (data->screen.sel != NULL && data->rectflag)
 		px = screen_size_x(&data->screen);
 	else
 		px = window_copy_find_length(wp, py);
@@ -2014,7 +2057,7 @@ window_copy_cursor_up(struct window_pane *wp, int scroll_only)
 		data->lastsx = ox;
 	}
 
-	if (s->sel.lineflag == LINE_SEL_LEFT_RIGHT && oy == data->sely)
+	if (data->lineflag == LINE_SEL_LEFT_RIGHT && oy == data->sely)
 		window_copy_other_end(wp);
 
 	data->cx = data->lastcx;
@@ -2036,7 +2079,7 @@ window_copy_cursor_up(struct window_pane *wp, int scroll_only)
 		}
 	}
 
-	if (!data->screen.sel.flag || !data->rectflag) {
+	if (data->screen.sel != NULL || !data->rectflag) {
 		py = screen_hsize(data->backing) + data->cy - data->oy;
 		px = window_copy_find_length(wp, py);
 		if ((data->cx >= data->lastsx && data->cx != px) ||
@@ -2044,9 +2087,9 @@ window_copy_cursor_up(struct window_pane *wp, int scroll_only)
 			window_copy_cursor_end_of_line(wp);
 	}
 
-	if (s->sel.lineflag == LINE_SEL_LEFT_RIGHT)
+	if (data->lineflag == LINE_SEL_LEFT_RIGHT)
 		window_copy_cursor_end_of_line(wp);
-	else if (s->sel.lineflag == LINE_SEL_RIGHT_LEFT)
+	else if (data->lineflag == LINE_SEL_RIGHT_LEFT)
 		window_copy_cursor_start_of_line(wp);
 }
 
@@ -2064,7 +2107,7 @@ window_copy_cursor_down(struct window_pane *wp, int scroll_only)
 		data->lastsx = ox;
 	}
 
-	if (s->sel.lineflag == LINE_SEL_RIGHT_LEFT && oy == data->endsely)
+	if (data->lineflag == LINE_SEL_RIGHT_LEFT && oy == data->endsely)
 		window_copy_other_end(wp);
 
 	data->cx = data->lastcx;
@@ -2078,7 +2121,7 @@ window_copy_cursor_down(struct window_pane *wp, int scroll_only)
 			window_copy_redraw_lines(wp, data->cy - 1, 2);
 	}
 
-	if (!data->screen.sel.flag || !data->rectflag) {
+	if (data->screen.sel == NULL || !data->rectflag) {
 		py = screen_hsize(data->backing) + data->cy - data->oy;
 		px = window_copy_find_length(wp, py);
 		if ((data->cx >= data->lastsx && data->cx != px) ||
@@ -2086,9 +2129,9 @@ window_copy_cursor_down(struct window_pane *wp, int scroll_only)
 			window_copy_cursor_end_of_line(wp);
 	}
 
-	if (s->sel.lineflag == LINE_SEL_LEFT_RIGHT)
+	if (data->lineflag == LINE_SEL_LEFT_RIGHT)
 		window_copy_cursor_end_of_line(wp);
-	else if (s->sel.lineflag == LINE_SEL_RIGHT_LEFT)
+	else if (data->lineflag == LINE_SEL_RIGHT_LEFT)
 		window_copy_cursor_start_of_line(wp);
 }
 
@@ -2147,14 +2190,14 @@ window_copy_cursor_jump_back(struct window_pane *wp)
 }
 
 static void
-window_copy_cursor_jump_to(struct window_pane *wp, int jump_again)
+window_copy_cursor_jump_to(struct window_pane *wp)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*back_s = data->backing;
 	struct grid_cell		 gc;
 	u_int				 px, py, xx;
 
-	px = data->cx + 1 + jump_again;
+	px = data->cx + 2;
 	py = screen_hsize(back_s) + data->cy - data->oy;
 	xx = window_copy_find_length(wp, py);
 
@@ -2172,7 +2215,7 @@ window_copy_cursor_jump_to(struct window_pane *wp, int jump_again)
 }
 
 static void
-window_copy_cursor_jump_to_back(struct window_pane *wp, int jump_again)
+window_copy_cursor_jump_to_back(struct window_pane *wp)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*back_s = data->backing;
@@ -2185,7 +2228,7 @@ window_copy_cursor_jump_to_back(struct window_pane *wp, int jump_again)
 	if (px > 0)
 		px--;
 
-	if (jump_again && px > 0)
+	if (px > 0)
 		px--;
 
 	for (;;) {
@@ -2359,7 +2402,7 @@ window_copy_scroll_up(struct window_pane *wp, u_int ny)
 		window_copy_write_line(wp, &ctx, 1);
 	if (screen_size_y(s) > 3)
 		window_copy_write_line(wp, &ctx, screen_size_y(s) - 2);
-	if (s->sel.flag && screen_size_y(s) > ny)
+	if (s->sel != NULL && screen_size_y(s) > ny)
 		window_copy_write_line(wp, &ctx, screen_size_y(s) - ny - 1);
 	screen_write_cursormove(&ctx, data->cx, data->cy);
 	screen_write_stop(&ctx);
@@ -2387,7 +2430,7 @@ window_copy_scroll_down(struct window_pane *wp, u_int ny)
 	screen_write_cursormove(&ctx, 0, 0);
 	screen_write_insertline(&ctx, ny, 8);
 	window_copy_write_lines(wp, &ctx, 0, ny);
-	if (s->sel.flag && screen_size_y(s) > ny)
+	if (s->sel != NULL && screen_size_y(s) > ny)
 		window_copy_write_line(wp, &ctx, ny);
 	else if (ny == 1) /* nuke position */
 		window_copy_write_line(wp, &ctx, 1);
@@ -2395,14 +2438,17 @@ window_copy_scroll_down(struct window_pane *wp, u_int ny)
 	screen_write_stop(&ctx);
 }
 
-int
-window_copy_scroll_position(struct window_pane *wp)
+void
+window_copy_add_formats(struct window_pane *wp, struct format_tree *ft)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 
 	if (wp->mode != &window_copy_mode)
-		return (-1);
-	return (data->oy);
+		return;
+
+	format_add(ft, "selection_present", "%d", data->screen.sel != NULL);
+	format_add(ft, "scroll_position", "%d", data->oy);
+	format_add(ft, "rectangle_toggle", "%d", data->rectflag);
 }
 
 static void
@@ -2432,7 +2478,7 @@ window_copy_move_mouse(struct mouse_event *m)
 	if (wp == NULL || wp->mode != &window_copy_mode)
 		return;
 
-	if (cmd_mouse_at(wp, m, &x, &y, 1) != 0)
+	if (cmd_mouse_at(wp, m, &x, &y, 0) != 0)
 		return;
 
 	window_copy_update_cursor(wp, x, y);
@@ -2482,16 +2528,3 @@ window_copy_drag_update(__unused struct client *c, struct mouse_event *m)
 	if (window_copy_update_selection(wp, 1))
 		window_copy_redraw_selection(wp, old_cy);
 }
-
-const char *
-window_copy_search_string(struct window_pane *wp)
-{
-	struct window_copy_mode_data	*data;
-
-	if (wp->mode != &window_copy_mode)
-		return ("");
-	data = wp->modedata;
-	if (data->searchtype == WINDOW_COPY_OFF || data->searchstr == NULL)
-		return ("");
-	return (data->searchstr);
-}
